/* FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  Copyright (C) 1998 BERO 
 *  Copyright (C) 2002 Ben Parnell
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"
#include "x6502.h"
#include "fce.h"
#define INESPRIV
#include "ines.h"
#include "svga.h"
//#include "state.h"
#include "memory.h"
#include "cart.h"

#include "saturn/shared.h"

extern byte *base_rom;
//extern byte base_rom[0x20000];
uint8 *ROM=NULL;
uint8 *VROM=NULL;

uint32 ROM_size;
uint32 VROM_size;

static int MMC_init(int type);
void (*MapClose)(void);
void (*MapperReset)(void);

static int MapperNo;
//static int SaveGame=0;

static iNES_HEADER head;

/*  MapperReset() is called when the NES is reset(with the reset button).  
    Mapperxxx_init is called when the NES has been powered on.
*/

static void iNESGI(int h)
{
    switch(h)
    {
        case GI_RESETM2:
            if(MapperReset)
                MapperReset();
            break;

        case GI_POWER:
            SetReadHandler(0x8000,0xFFFF,CartBR);
            MMC_init(MapperNo);
            break;

        case GI_CLOSE:
            if(ROM) {free(ROM);ROM=0;}
            if(VROM) {free(VROM);VROM=0;}
            if(MapClose) MapClose();
            break;
    }
}

uint32 uppow2(uint32 n)
{
    int x;
    
    for(x=31;x>=0;x--)
        if(n&(1<<x))
        {
            if((1<<x)!=n)
                return(1<<(x+1));
            break;
        }

    return n;
}

int iNESLoad(char *name,  uint8 *rom)
{
    int i;
    for (i=0;i < 4 ; i++)
		head.ID[i]=rom[i];

	head.ROM_size  = rom[4];
	head.VROM_size = rom[5];
	head.ROM_type  = rom[6];
	head.ROM_type2 = rom[7];
	
	for (i=0;i < 8 ; i++)
		head.reserve[i]=rom[i+8];
	  	
    if(memcmp(&head,"NES\x1a",4))
        return 0;
    
    if(!head.ROM_size)
        head.ROM_size++;

    ROM_size    = head.ROM_size;
    ROM_size    = uppow2(ROM_size);
    VROM_size   = head.VROM_size;
    VROM_size   = uppow2(VROM_size);

    MapperNo= (head.ROM_type>>4);
    MapperNo|=(head.ROM_type2&0xF0);

    Mirroring = (head.ROM_type&1);
	if(head.ROM_type&8) Mirroring=2;

//	ROM= (uint8 *) 0x00202000+16;
	ROM= (uint8 *) rom+16;

	ResetCartMapping();
	SetupCartPRGMapping(0,ROM,ROM_size*0x4000,0);
    SetupCartPRGMapping(1,WRAM,8192,1);

   if(VROM_size)
       VROM=(uint8 *) rom+16+0x4000*ROM_size;
//       VROM=(uint8 *) 0x00202000+16+0x4000*ROM_size;

    FCEU_printf(" PRG ROM:  %3d x 16k\n CHR ROM:  %3d x  8k\n Mapper:  %d\n Mirroring: %s\n",head.ROM_size,head.VROM_size,MapperNo,Mirroring==2?"None(Four-screen)":Mirroring?"Vertical":"Horizontal");
//    if(head.ROM_type&2) FCEU_printf(" Battery-backed.\n");
//    if(head.ROM_type&4) FCEU_printf(" Trained.\n");

	if(VROM_size)
         SetupCartCHRMapping(0,VROM,VROM_size*0x2000,0);

	if(Mirroring>=0x10) 
        SetupCartMirroring(2+(Mirroring&1),1,0);
	else
         SetupCartMirroring(Mirroring&1,(Mirroring&4)>>2,0);

    GameInterface=iNESGI;
    return 1;
}

#include	"banksw.h"

void FASTAPASS(1) onemir(uint8 V)
{
    if(Mirroring==2) return;
        if(V>1)
            V=1;

	Mirroring=0x10|V;
	setmirror(MI_0+V);
}

void FASTAPASS(1) MIRROR_SET2(uint8 V)
{
	if(Mirroring==2) return;
	Mirroring=V;
	setmirror(V);
}

void FASTAPASS(1) MIRROR_SET(uint8 V)
{
	if(Mirroring==2) return;
	V^=1;
	Mirroring=V;
	setmirror(V);
}

static void NONE_init(void)
{
    ROM_BANK16(0x8000,0);
    ROM_BANK16(0xC000,~0);
    
    if(VROM_size) 
    VROM_BANK8(0);
//        else
//	 setvram8(CHRRAM);
}

void (*MapInitTab[256])(void)=
{
	0,
	Mapper1_init,Mapper2_init,Mapper3_init,Mapper4_init,
	Mapper5_init,Mapper6_init,Mapper7_init,Mapper8_init,
	Mapper9_init,Mapper10_init,Mapper11_init,0,
	Mapper13_init,0,Mapper15_init,Mapper16_init,
	Mapper17_init,Mapper18_init,Mapper19_init,0,
	Mapper21_init,Mapper22_init,Mapper23_init,Mapper24_init,
	Mapper25_init,Mapper26_init,0,0,
	0,0,0,Mapper32_init,
	Mapper33_init,Mapper34_init,0,0,
	0,0,0,Mapper40_init,
	Mapper41_init,Mapper42_init,Mapper43_init,Mapper44_init,
	Mapper45_init,Mapper46_init,Mapper47_init,Mapper33_init,Mapper49_init,0,Mapper51_init,Mapper52_init,
	0,0,0,0,Mapper57_init,0,Mapper59_init,0,
	Mapper61_init,0,0,Mapper64_init,
	Mapper65_init,Mapper66_init,Mapper67_init,Mapper68_init,
	Mapper69_init,Mapper70_init,Mapper71_init,Mapper72_init,
	Mapper73_init,Mapper74_init,Mapper75_init,Mapper76_init,
	Mapper77_init,Mapper78_init,Mapper79_init,Mapper80_init,
	0,Mapper82_init,Mapper83_init,0,
	Mapper85_init,Mapper86_init,Mapper87_init,Mapper88_init,
	Mapper89_init,0,0,0,
	0,0,0,0,
	0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0
};

static DECLFW(BWRAM)
{
    WRAM[A-0x6000]=V;
}

static DECLFR(AWRAM)
{
    return WRAM[A-0x6000];
//return NULL;
}

void (*MapStateRestore)(int version);
void iNESStateRestore(int version)
{
 int x;
 
 if(!MapperNo) return;

 for(x=0;x<4;x++)
//  setprg8(0x8000+x*8192,PRGBankList[x]);

 if(VROM_size)
  for(x=0;x<8;x++) 
    setchr1(0x400*x,CHRBankList[x]);

 switch(Mirroring)
 {
   case 0:setmirror(MI_H);break;
   case 1:setmirror(MI_V);break;
   case 0x12:
   case 0x10:setmirror(MI_0);break;
   case 0x13:
   case 0x11:setmirror(MI_1);break;
 }
 if(MapStateRestore) MapStateRestore(version);
}

static int MMC_init(int type)
{
    int x;
    
    GameStateRestore=iNESStateRestore;
    MapClose=0;
	MapperReset=0;
    MapStateRestore=0;

    setprg8r(1,0x6000,0);

	// Fix this if the MMC3 iNES code ever sets the WRAM read/write handlers
	 //  itself.
	
    SetReadHandler(0x6000,0x7FFF,AWRAM);
    SetWriteHandler(0x6000,0x7FFF,BWRAM);
    
	// This statement represents atrocious code.  I need to rewrite
	//   all of the iNES mapper code... 
    IRQCount=IRQLatch=IRQa=0;

    if(head.ROM_type&2)
	{
        extern uint8 MMC5WRAMsize,MMC1WRAMsize;

        if(type==5 && MMC5WRAMsize==4)
            memset(GameMemBlock+32768,0,sizeof(GameMemBlock)-32768);
        else if(type==1 && MMC1WRAMsize==2)
        {
            memset(GameMemBlock,0,8192);
            memset(GameMemBlock+16384,0,sizeof(GameMemBlock)-16384);
        }
        else
            memset(GameMemBlock+8192,0,sizeof(GameMemBlock)-8192);
	}
	else
        memset(GameMemBlock,0,sizeof(GameMemBlock));

    NONE_init();
 
    if(type==5)
    {
        MMC5HackVROMMask=CHRmask4[0];
//          MMC5HackExNTARAMPtr=MapperExRAM+0x6000;
        MMC5Hack=1;
        MMC5HackVROMPTR=VROM;
        MMC5HackCHRMode=0;
    }
    if(MapInitTab[type]) 
        MapInitTab[type]();
    else if(type)
        FCEU_PrintError("iNES mapper #%d is not supported at all.",type);

    return 1;
}

